import os


class OIDC_PROVIDERS:
    GOOGLE = "google"
    OKTA = "okta"
    KEYCLOAK = "keycloak"
    CUSTOM = "custom"


# All the custom configurations required for an OIDC provider to work with Authlib
# will go in here. We'll pass these while registering a client.
_CONFIGS = {
    OIDC_PROVIDERS.GOOGLE: {
        "authorize_params": {"access_type": "offline", "prompt": "consent"}
    },
    OIDC_PROVIDERS.OKTA: {},
    OIDC_PROVIDERS.KEYCLOAK: {},
    OIDC_PROVIDERS.CUSTOM: {},
}


class BaseConfig(object):
    SECRET_KEY = os.environ.get("FLASK_OIDC_SECRET_KEY", "!-flask-oidc-secret-key")
    WHITELISTED_ENDPOINTS = os.environ.get(
        "FLASK_OIDC_WHITELISTED_ENDPOINTS", "status,healthcheck,health"
    )

    OIDC_PROVIDER = os.environ.get("FLASK_OIDC_PROVIDER_NAME", "keycloak")
    OIDC_SCOPES = os.environ.get("FLASK_OIDC_SCOPES", "openid email profile")
    USER_ID_FIELD = os.environ.get("FLASK_OIDC_USER_ID_FIELD", "sub")
    CLIENT_ID = os.environ.get("FLASK_OIDC_CLIENT_ID", "")
    CLIENT_SECRET = os.environ.get("FLASK_OIDC_CLIENT_SECRET", "")
    SCHEME = os.environ.get("FLASK_OIDC_FORCE_SCHEME", "http")
    REDIRECT_URI = os.environ.get("FLASK_OIDC_REDIRECT_URI", "/auth")
    OVERWRITE_REDIRECT_URI = os.environ.get("FLASK_OIDC_OVERWRITE_REDIRECT_URI", "/")
    CONFIG_URL = os.environ.get("FLASK_OIDC_CONFIG_URL", "")
    # OAuth role: client or server (resource server) In our case, client means that
    # authentication in done through Flask, using /login and /auth endpoints. Server means
    # that the app is authenticated by either passing a refresh token to the /token endpoint
    # or by passing an access token in your request.
    ROLE = os.environ.get("FLASK_OIDC_ROLE", "client")
    # Access type options for Keycloak: public/confidential/bearer-only
    ACCESS_TYPE = os.environ.get("FLASK_OIDC_ACCESS_TYPE", "confidential")
    # Userinfo keys to return in user object

    OIDC_PROVIDER_PARAMETERS_FILE = os.environ.get(
        "FLASK_OIDC_PROVIDER_ADDITIONAL_PARAMETERS_FILE_PATH", None
    )
    if OIDC_PROVIDER_PARAMETERS_FILE:
        import json

        with open(OIDC_PROVIDER_PARAMETERS_FILE) as parameters_file:
            parameters = json.load(parameters_file)
            _CONFIGS[OIDC_PROVIDER] = parameters

    SQLALCHEMY_TRACK_MODIFICATIONS = os.environ.get(
        "SQLALCHEMY_TRACK_MODIFICATIONS", False
    )
    SQLALCHEMY_DATABASE_URI = os.environ.get(
        "SQLALCHEMY_DATABASE_URI", "sqlite:///sessions.db"
    )

    INVALID_TOKEN_MESSAGE = "The access token used is invalid."
    NO_TOKEN_MESSAGE = "Use of resource server requires that you pass an access token."
    EXPIRED_TOKEN_MESSAGE = "The access token has expired."

    USERINFO_MAPPING = {
        "sub": "id",
        "preferred_username": "username",
        "email": "email",
        "name": "name",
        "given_name": "first_name",
        "family_name": "last_name",
        "email_verified": "email_verified",
    }



import logging
import time
from sqlalchemy import Column, Integer, String, TEXT
from flask import current_app as app

LOGGER = logging.getLogger(f"flaskoidc.{__name__}")


class OAuth2Token(app.db.Model):
    id = Column(Integer, primary_key=True)
    user_id = Column(String(320), nullable=False)
    name = Column(String(20), nullable=False)

    access_token = Column(TEXT, nullable=False)
    expires_in = Column(Integer, default=0)
    scope = Column(String(320), default=0)
    token_type = Column(String(20))
    refresh_token = Column(TEXT)
    expires_at = Column(Integer, default=0)

    def to_token(self):
        return dict(
            access_token=self.access_token,
            expires_in=self.expires_in,
            scope=self.scope,
            token_type=self.token_type,
            refresh_token=self.refresh_token,
            expires_at=self.expires_at,
        )

    @property
    def is_active(self):
        return self.expires_at > round(time.time())

    @staticmethod
    def save(**kwargs):
        OAuth2Token.delete(name=kwargs['name'], user_id=kwargs['user_id'])
        token = OAuth2Token(**OAuth2Token.filter_token(kwargs))
        app.db.session.add(token)
        app.db.session.commit()
        return token

    @staticmethod
    def get(**kwargs):
        return OAuth2Token.query.filter_by(**kwargs).first()

    @staticmethod
    def delete(**kwargs):
        OAuth2Token.query.filter_by(**kwargs).delete()
        app.db.session.commit()

    @staticmethod
    def all():
        return OAuth2Token.query.all()

    @staticmethod
    def filter_token(token):
        return {_key: token.get(_key) for _key in token.keys() if _key in OAuth2Token.__dict__.keys()}



import logging
from flask import session, request, _app_ctx_stack
from werkzeug.exceptions import BadRequest
from authlib.integrations.flask_client import FlaskRemoteApp
from authlib.oidc.core.errors import LoginRequiredError
from flaskoidc.models import OAuth2Token

LOGGER = logging.getLogger(__name__)


class FlaskAuthClient(FlaskRemoteApp):
    """Extension of FlaskRemoteApp, the default client class used by Flask OAuth.
    self.auth_client seen in __init__.py is an instance of this class. This class
    implements additional checking for token expiration and automatic token refresh.
    """

    @property
    def token(self):
        """Extended implementation of the FlaskRemoteApp token property. First checks
        if a token has already been set. If not, collects a token from the database and
        refreshes if expired.
        """
        ctx = _app_ctx_stack.top
        attr = 'authlib_oauth_token_{}'.format(self.name)
        token = getattr(ctx, attr, None)
        if token:
            return token

        if session.get("user"):
            token = OAuth2Token.get(name=self.name, user_id=session["user"]["id"])
            if not token:
                raise LoginRequiredError("")
        else:
            access_token = self.parse_access_token()
            if not access_token:
                raise LoginRequiredError(
                    "No active user session and no access token provided.")
            token = OAuth2Token.get(access_token=access_token)
            if not token:
                raise LoginRequiredError("Access token is invalid.")
        if token.is_active:
            return token
        else:
            new_token = self.refresh_token(token.refresh_token)
            return OAuth2Token.save(name=self.name,
                                    user_id=token.user_id,
                                    **new_token)

    @token.setter
    def token(self, token):
        ctx = _app_ctx_stack.top
        attr = 'authlib_oauth_token_{}'.format(self.name)
        setattr(ctx, attr, token)

    def refresh_token(self, refresh_token=None):
        """Calls the refresh token endpoint to receive a new auth token.
        """
        metadata = self.load_server_metadata()
        with self._get_oauth_client(**metadata) as client:
            new_token = client.refresh_token(metadata['token_endpoint'],
                                             refresh_token=refresh_token)
            if new_token.get('error') == 'invalid_grant':
                raise LoginRequiredError('Refresh token is expired.')
            return new_token

    def handle_auth_token(self, token, userinfo_mapping=None):
        """Convenience method used post-authentication when an auth token is received
        from the auth provider. Parses token, saves it in the database, and initializes
        user session.
        """
        LOGGER.debug(f"Token Info: {token}")
        user = self.parse_id_token(token)
        LOGGER.debug(f"User Info: {user}")
        if userinfo_mapping:
            user = {user_key: user.get(userinfo_key)
                    for userinfo_key, user_key in userinfo_mapping.items()}
        OAuth2Token.save(name=self.name, user_id=user["id"], **token)
        session["user"] = user
        session.permanent = True
        return user

    @staticmethod
    def parse_access_token():
        if "Authorization" in request.headers and \
                request.headers["Authorization"].startswith("Bearer "):
            return request.headers["Authorization"].split(None, 1)[1].strip()
        return None



import logging
from authlib.integrations.flask_client import OAuth
from authlib.oidc.core.errors import LoginRequiredError
from flask import Flask, request, session, jsonify, abort, redirect
from flask.helpers import get_env, get_debug_flag, make_response, url_for
from flask_sqlalchemy import SQLAlchemy
from flaskoidc.config import BaseConfig, _CONFIGS

LOGGER = logging.getLogger(__name__)


class FlaskOIDC(Flask):
    def _before_request(self):
        # Whitelisted Endpoints i.e., health checks and status url
        whitelisted_endpoints = self.config.get("WHITELISTED_ENDPOINTS")
        LOGGER.debug(f"Whitelisted Endpoint: {whitelisted_endpoints}")

        # Add auth endpoints to whitelisted endpoint as well, so not to check for token on that
        whitelisted_endpoints += (
            f",login,logout,token,{self.config.get('REDIRECT_URI').strip('/')}"
        )

        if request.path.strip("/") in whitelisted_endpoints.split(",") or \
                request.endpoint in whitelisted_endpoints.split(","):
            return

        try:
            assert self.auth_client.token
        except LoginRequiredError as e:
            if self.config.get("ROLE") == "client":
                LOGGER.exception(
                    "User not logged in, redirecting to auth", exc_info=True
                )
                return redirect(url_for("logout", _external=True))
            else:
                abort(make_response(jsonify({"message": e.description}), 401))

    def __init__(self, *args, **kwargs):
        super(FlaskOIDC, self).__init__(*args, **kwargs)

        self.db = SQLAlchemy(self)
        _provider = self.config.get("OIDC_PROVIDER").lower()

        if _provider not in _CONFIGS.keys():
            LOGGER.info(
                f"""
            [flaskoidc Notice] I have not verified the OIDC Provider that you have 
            selected i.e., "{_provider}" with this package yet. 
            If you encounter any issue while using this library with "{_provider}",
            please do not hesitate to create an issue on Github. (https://github.com/verdan/flaskoidc)
            """
            )

        with self.app_context():
            from flaskoidc.auth_client import FlaskAuthClient

            self.db.create_all()

            oauth = OAuth(self)

            self.auth_client = oauth.register(
                name=_provider,
                client_cls=FlaskAuthClient,
                server_metadata_url=self.config.get("CONFIG_URL"),
                client_kwargs={
                    "scope": self.config.get("OIDC_SCOPES"),
                },
                **_CONFIGS.get(_provider) if _CONFIGS.get(_provider) else {},
            )

        # Register the before request function that will make sure each
        # request is authenticated before processing
        self.before_request(self._before_request)

        @self.route("/login")
        def login():
            redirect_uri = url_for("auth", _external=True, _scheme=self.config.get("SCHEME"))
            return self.auth_client.authorize_redirect(redirect_uri)

        @self.route(self.config.get("REDIRECT_URI"))
        def auth():
            try:
                token = self.auth_client.authorize_access_token()
                self.auth_client.handle_auth_token(token)
                return redirect(self.config.get("OVERWRITE_REDIRECT_URI"))
            except Exception as ex:
                LOGGER.exception(ex)
                raise ex

        @self.route("/token", methods=["GET", "POST"])
        def token():
            if request.method == "GET":
                access_token = self.auth_client.token.access_token
                return make_response(jsonify({"access_token": access_token}), 200)
            elif request.method == "POST":
                data = request.form.to_dict()
                token = self.auth_client.refresh_token(data["refresh_token"])
                # fixes 'missing nonce' which shouldn't happen in production:
                # session.pop("_keycloak_authlib_nonce_")
                user = self.auth_client.handle_auth_token(token, self.config.get('USERINFO_MAPPING'))
                user_details_keys = self.config.get('USER_DETAILS_KEYS', '').split(',')
                user_details = {key: val for key, val in user.items()
                                if key in user_details_keys}
                return make_response(jsonify(user_details), 200)

        @self.route("/logout")
        def logout():
            # ToDo: Think of if we should delete the session entity or not
            # if session.get("user"):
            #     OAuth2Token.delete(name=_provider, user_id=session["user"]["id"])
            session.pop("user")
            if self.config.get("ROLE") == "client":
                return redirect(url_for("login"))    
            return make_response(jsonify({"message": "Logged out."}), 200)

    def make_config(self, instance_relative=False):
        """
        Overriding the default `make_config` function in order to support
        Flask OIDC package and all of their settings.
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        defaults = dict(self.default_config)
        defaults["ENV"] = get_env()
        defaults["DEBUG"] = get_debug_flag()

        _required_fields = ["CLIENT_ID", "CLIENT_SECRET", "CONFIG_URL"]

        # Append all the configurations from the base config class.
        for key, value in BaseConfig.__dict__.items():
            if not key.startswith("__"):
                if key in ["CLIENT_ID", "CLIENT_SECRET"]:
                    key = f"{BaseConfig.OIDC_PROVIDER.upper()}_{key}"

                if key in _required_fields and not value:
                    raise RuntimeError(
                        f"Invalid Configuration: {key} is required and can not be empty."
                    )

                defaults[key] = value
        return self.config_class(root_path, defaults)



from os import path

from setuptools import setup, find_packages

requirements = [
    "Flask>=1.0.2",
    "Authlib==0.15.4",
    "requests==2.25.1",
    "Flask-SQLAlchemy==2.5.1",
]

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

__version__ = "1.0.3"

setup(
    name="hey-auth",
    version=__version__,
    description="Flask wrapper with pre-configured OAuth2 and OIDC support",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/coderbarns/flaskoidc.git",
    author="Aarni Alasaarela",
    author_email="coderbarns@gmail.com",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    dependency_links=[],
    install_requires=requirements,
    python_requires=">=3.6",
)
